----------------------------
--      Constants         --
----------------------------

local s_strLowHeathWavPath = "Interface\\AddOns\\SRNAnubHealerMod\\klaxon1.wav";
local s_strPenetratingColdWavPath = "Interface\\AddOns\\SRNAnubHealerMod\\incoming.wav";


local s_rgstrIconEncodings =
{
	[1] = "{star}",
	[2] = "{circle}",
	[3] = "{diamond}",
	[4] = "{triangle}",
	[5] = "{moon}",
	[6] = "{square}",
	[7] = "{cross}",
	[8] = "{skull}",
};

local s_rgiHPPerSecond =
{
	[1]  = 250,
	[2]  = 500,
	[3]  = 750,
	[4]  = 1000,
	[5]  = 1250,
	[6]  = 1562,
	[7]  = 1953,
	[8]  = 2442,
	[9]  = 3053,
	[10] = 3817,
	[11] = 4772,
	[12] = 5965,
	[13] = 7457,
	[14] = 9321,
	[15] = 11652,
};
local s_iMaxComputedSecond = 15;

local s_colorDefaultBG	= TOOLTIP_DEFAULT_BACKGROUND_COLOR;
local s_colorYellowBG	= { r = 208, g = 191, b = 0 };
local s_colorRedBG	= { r = 200, g = 0, b = 0 };

local options = {
	type = "group",
	name = "SRNAnubHealerMod",
	get = function(info) return SRNAnubHealerMod.db.profile[ info[#info] ] end,
	set = function(info, value) SRNAnubHealerMod.db.profile[ info[#info] ] = value end,
	args = {
		General = {
			order = 1,
			type = "group",
			name = "General Settings",
			desc = "General Settings",
			args = {
				lblIntro = {
					order = 1,
					type = "description",
					name = "SRNAnubHealerMod is a mod for Anub'arak in Trial of the Grand Crusader.",
				},
				PenetratingCold = {
					type = "group",
					name = "Penetrating Cold",
					guiInline = true,
					order = 2,
					args = {
						fEnablePCSound = {
							type = "toggle",
							name = "Enable warning sound",
							desc = "Allow sound warning that Anub'arak has casted Penetrating Cold.",
							order = 1,
						},
						btnTestPCSound = {
							type = "execute",
							name = "Preview sound",
							desc = "Previews the sound warning for Penetrating Cold.",
							order = 2,
							func = function() PlaySoundFile(s_strPenetratingColdWavPath) end,
						};
					},
				},
				RaidHealth = {
					type = "group",
					name = "Raid Health",
					guiInline = true,
					order = 3,
					args = {
						lblIntro = {
							type = "description",
							name = "SRNAnubHealerMod monitors the health of the raid and reports the "
								.. "number of seconds remaining before someone dies to Leeching Swarm "
								.. "(assuming no additional healing or damage).  If your JoL and "
								.. "Healing Stream Totems are set up correctly, this should hover around "
								.. "4-6 seconds.",
							order = 1,
						},
						fEnableLowHeathSound = {
							type = "toggle",
							name = "Enable warning sound",
							desc = "Allow sound warning when someone is at low health.",
							order = 2,
						},
						btnTestLowHeathSound = {
							type = "execute",
							name = "Preview sound",
							desc = "Previews the sound alert for a raid member at low health.",
							order = 3,
							func = function() PlaySoundFile(s_strLowHeathWavPath) end,
						};
						iLowHealthSoundLevel = {
							type = "range",
							name = "Warning sound threshold (seconds)",
							desc = "If the raid member with lowest health has less than this many seconds to live, sound an alert.",
							order = 4,
							width = "double",
							min = 0,
							max = s_iMaxComputedSecond - 1,
							step = 1,
							disabled = function() return not SRNAnubHealerMod.db.profile.fEnableLowHeathSound end,
						},
						fEnableLowHealthFrame = {
							type = "toggle",
							name = "Enable frame",
							desc = "Show the SRNAnubHealerMod frame.  The frame will be visible only in Trial of the Crusader.",
							order = 5,
							set = function(info, value)
									SRNAnubHealerMod.db.profile.fEnableLowHealthFrame = value;
									SRNAnubHealerMod:CheckFrameState();
								end,
						},
						btnResetLowHealthFrameLocation = {
							type = "execute",
							name = "Reset frame",
							desc = "Recenters the SRNAnubHealerMod frame.",
							order = 6,
							func = function() SRNAnubHealerMod:ResetFrameLocation() end,
							disabled = function() return not SRNAnubHealerMod.db.profile.fEnableLowHealthFrame end,
						};
					},
				},
				PCDeathBlame = {
					type = "group",
					name = "Death Blaming",
					guiInline = true,
					order = 4,
					args = {
						lblIntro = {
							type = "description",
							name = "SRNAnubHealerMod can announce blame for Penetrating Cold deaths.  Assumes another mod, such as DBM, is marking the Penetrating Cold targets.",
							order = 1,
						},
						fEnableBlame = {
							type = "toggle",
							name = "Enable announce",
							desc = "Allow announcement of blame for Penetrating Cold deaths.",
							order = 2,
						},
						fBlameOnlyWhenAssist = {
							type = "toggle",
							name = "Only when assist",
							desc = "Only announce blame when you are Raid Leader or a Raid Assistant.",
							order = 3,
							disabled = function() return not SRNAnubHealerMod.db.profile.fEnableBlame end,
						},
						strBlameChannel = {
							type = "input",
							name = "Announce channel",
							desc = "The channel to announce blame.  Can also be \"SAY\", \"YELL\", \"PARTY\", \"RAID\", \"RAID_WARNING\", \"OFFICER\", or \"GUILD\".",
							order = 4,
							width = "double",
							disabled = function() return not SRNAnubHealerMod.db.profile.fEnableBlame end,
						},
						btnTestAnnounce = {
							type = "execute",
							name = "Test Announce",
							desc = "Sends a test message to the channel.",
							order = 5,
							func = function() SRNAnubHealerMod:AnnounceMessage("SRNAnubHealerMod: Test message") end,
							disabled = function() return not SRNAnubHealerMod.db.profile.fEnableBlame end,
						},
					},
				},
			},
		},
	},
};

------------------------------
--      Initialization      --
------------------------------

SRNAnubHealerMod = AceLibrary("AceAddon-3.0"):NewAddon("SRNAnubHealerMod", "AceEvent-3.0", "AceConsole-3.0")

function SRNAnubHealerMod:OnInitialize()
	SRNAnubHealerMod.rgiRaidIcon = {};
	SRNAnubHealerMod.rgiRaidIconLast = {};
	SRNAnubHealerMod.rgtmRaidIconLastTime = {};
	SRNAnubHealerMod.rgsAlternateCauses = {};
	SRNAnubHealerMod.tmLastLeech = 0;
	SRNAnubHealerMod.tmLastScan = 0;
	SRNAnubHealerMod.tmLastKlaxon = 0;
	SRNAnubHealerMod.tmPhase3Start = 0;
	SRNAnubHealerMod.fActive = false;

	-- Build the default settings array
	local DEFAULTS = {
		profile = {
			fEnablePCSound = true,
			fEnableLowHeathSound = true,
			iLowHealthSoundLevel = 4,
			fEnableLowHealthFrame = true,
			fEnableBlame = true,
			fBlameOnlyWhenAssist = true,
			strBlameChannel = "RAID",
		}
	};
	SRNAnubHealerMod.db = LibStub("AceDB-3.0"):New("SRNAnubHealerModDB", DEFAULTS, "Default")

	options.args.Profile = LibStub("AceDBOptions-3.0"):GetOptionsTable(SRNAnubHealerMod.db)
	LibStub("AceConfig-3.0"):RegisterOptionsTable("SRNAnubHealerMod", options)
	LibStub("AceConfigDialog-3.0"):SetDefaultSize("SRNAnubHealerMod", 640, 480)
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SRNAnubHealerMod", nil, nil, "General")
	LibStub("AceConfigDialog-3.0"):AddToBlizOptions("SRNAnubHealerMod", "Profile", "SRNAnubHealerMod", "Profile")
	SRNAnubHealerMod:RegisterChatCommand("ahm", function() LibStub("AceConfigDialog-3.0"):Open("SRNAnubHealerMod") end)
end


function SRNAnubHealerMod:OnEnable()
	SRNAnubHealerMod.fEnabled = true;
	SRNAnubHealerMod:RegisterEvent("ZONE_CHANGED_NEW_AREA");
	SRNAnubHealerMod:CheckActiveState();
end

function SRNAnubHealerMod:OnDisable()
	SRNAnubHealerMod.fEnabled = false;
	SRNAnubHealerMod:UnregisterEvent("ZONE_CHANGED_NEW_AREA");
	SRNAnubHealerMod:CheckActiveState();
end

------------------------------
--        Functions         --
------------------------------

---------------------------------------------------------------------------------------------------
--	SRNAnubHealerMod:CheckActiveState
--
--		Activates/deactivates the mod based on zone and enabled state
--
function SRNAnubHealerMod:CheckActiveState()

	--SRNAnubHealerMod:PrintMessage("Zone = " .. GetRealZoneText()); -- DEBUG

	if (SRNAnubHealerMod.fEnabled and ("Trial of the Crusader" == GetRealZoneText()
			--or "Icecrown" ==  GetRealZoneText() -- DEBUG
			)) then

		if (not SRNAnubHealerMod.fActive) then
			SRNAnubHealerMod.fActive = true;
			SRNAnubHealerMod:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
			SRNAnubHealerMod:PrintMessage("Add-on activated");
		end

	else

		if (SRNAnubHealerMod.fActive) then
			SRNAnubHealerMod.fActive = false;
			SRNAnubHealerMod:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED");
			SRNAnubHealerMod:PrintMessage("Add-on deactivated");
		end

	end

	SRNAnubHealerMod:CheckFrameState();

end

---------------------------------------------------------------------------------------------------
--	SRNAnubHealerMod:CheckFrameState
--
--		Activates/deactivates the mod based on zone and enabled state
--
function SRNAnubHealerMod:CheckFrameState()

	if (SRNAnubHealerMod.fActive and SRNAnubHealerMod.db.profile.fEnableLowHealthFrame) then

		--SRNAnubHealerMod:PrintMessage("Showing frame"); -- DEBUG

		-- Create Frame
		if (not SRNAnubHealerMod.Frame) then
			SRNAnubHealerMod.Frame = CreateFrame("Button", "SRNAnubHealerModFrame", UIParent);
			SRNAnubHealerMod.Frame:SetFrameStrata("MEDIUM");

			if (SRNAnubHealerMod.db.profile.sLowHealthFrameLocation) then
				SRNAnubHealerMod.Frame:SetPoint(SRNAnubHealerMod.db.profile.sLowHealthFrameLocation.strAnchor,
					UIParent,
					SRNAnubHealerMod.db.profile.sLowHealthFrameLocation.strParentAnchor,
					SRNAnubHealerMod.db.profile.sLowHealthFrameLocation.iOffsetX,
					SRNAnubHealerMod.db.profile.sLowHealthFrameLocation.iOffsetY
					);
			else
				SRNAnubHealerMod.Frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
			end

			SRNAnubHealerMod.Frame:SetHeight(28);
			SRNAnubHealerMod.Frame:SetWidth(180);
			SRNAnubHealerMod.Frame:SetMovable(true);
			SRNAnubHealerMod.Frame:EnableMouse(true);
			SRNAnubHealerMod.Frame:SetScript("OnMouseDown", function() SRNAnubHealerMod.Frame:StartMoving(); end);
			SRNAnubHealerMod.Frame:SetScript("OnMouseUp", function() SRNAnubHealerMod.Frame:StopMovingOrSizing(); SRNAnubHealerMod:PersistFrameLocation(); end);
			SRNAnubHealerMod.Frame:SetScript("OnClick", function() LibStub("AceConfigDialog-3.0"):Open("SRNAnubHealerMod"); end);
			SRNAnubHealerMod.Frame:RegisterForClicks("RightButtonUp");

			SRNAnubHealerMod.Frame:SetBackdrop(
				{
					bgFile = "Interface/Tooltips/UI-Tooltip-Background", 
					edgeFile = "Interface/Tooltips/UI-Tooltip-Border", 
					tile = true, tileSize = 16, edgeSize = 16, 
					insets = { left = 4, right = 4, top = 4, bottom = 4 }
				});

			SRNAnubHealerMod.Frame:SetBackdropColor(0, 0, 0, 1);
			SRNAnubHealerMod.Frame:SetBackdropBorderColor(1, 1, 1, 1);
		end

		-- Create Title text
		if (not SRNAnubHealerMod.FrameTitle) then
			SRNAnubHealerMod.FrameTitle = SRNAnubHealerMod.Frame:CreateFontString(nil, nil, "GameFontNormalSmall");
			SRNAnubHealerMod.FrameTitle:SetPoint("BOTTOMLEFT", SRNAnubHealerMod.Frame, "TOPLEFT", 4, 2);
			SRNAnubHealerMod.FrameTitle:SetJustifyH("LEFT");
			SRNAnubHealerMod.FrameTitle:SetTextColor(0.5, 1, 0.5, 1);
			SRNAnubHealerMod.FrameTitle:SetText("SRNAnubHealerMod");
		end

		-- Create body text
		if (not SRNAnubHealerMod.FrameMessage) then
			SRNAnubHealerMod.FrameMessage = SRNAnubHealerMod.Frame:CreateFontString(nil, nil, "GameFontNormal");
			SRNAnubHealerMod.FrameMessage:SetPoint("LEFT", SRNAnubHealerMod.Frame, "LEFT", 6, 0);
			SRNAnubHealerMod.FrameMessage:SetJustifyH("LEFT");
			SRNAnubHealerMod.FrameMessage:SetTextColor(1, 1, 1, 1);
			SRNAnubHealerMod.FrameMessage:SetText("Initializing");
		end

		SRNAnubHealerMod.Frame:Show();

	else

		--SRNAnubHealerMod:PrintMessage("Hiding frame"); -- DEBUG

		if (SRNAnubHealerMod.Frame) then
			SRNAnubHealerMod.PersistFrameLocation();
			SRNAnubHealerMod.Frame:Hide();
		end

	end

end

---------------------------------------------------------------------------------------------------
--	SRNAnubHealerMod:ResetFrameLocation
--
--		Resets the frame location (and persists it).
--
function SRNAnubHealerMod:ResetFrameLocation()

	SRNAnubHealerMod.db.profile.sLowHealthFrameLocation = nil;

	if (SRNAnubHealerMod.Frame) then
		SRNAnubHealerMod.Frame:SetPoint("CENTER", UIParent, "CENTER", 0, 0);
	end

end

---------------------------------------------------------------------------------------------------
--	SRNAnubHealerMod:PersistFrameLocation
--
--		Prints a message, only if the options permit it.
--
function SRNAnubHealerMod:PersistFrameLocation()

	if (SRNAnubHealerMod.Frame) then
		local strAnchor, _, strParentAnchor, iOffsetX, iOffsetY = SRNAnubHealerMod.Frame:GetPoint();

		SRNAnubHealerMod.db.profile.sLowHealthFrameLocation =
			{
			strAnchor = strAnchor,
			strParentAnchor = strParentAnchor,
			iOffsetX = iOffsetX,
			iOffsetY = iOffsetY,
			};
	end

end


---------------------------------------------------------------------------------------------------
--	SRNAnubHealerMod:PrintMessage
--
--		Prints a message, only if the options permit it.
--
function SRNAnubHealerMod:PrintMessage(strMsg)

	DEFAULT_CHAT_FRAME:AddMessage("|cff7fff7fAnubHealerMod|r: " .. strMsg);

end

---------------------------------------------------------------------------------------------------
--	SRNAnubHealerMod:AnnounceMessage
--
--		Announces a message to the user-specified channel
--
function SRNAnubHealerMod:AnnounceMessage(strMsg)

	local strChannelUpper = SRNAnubHealerMod.db.profile.strBlameChannel:upper();

	if ("SAY" == strChannelUpper
			or "YELL" == strChannelUpper
			or "PARTY" == strChannelUpper
			or "GUILD" == strChannelUpper
			or "OFFICER" == strChannelUpper
			or "RAID" == strChannelUpper
			or "RAID_WARNING" == strChannelUpper) then
		SendChatMessage(strMsg, strChannelUpper);
	else
		SendChatMessage(strMsg, "CHANNEL", nil, GetChannelName(strChannelUpper));
	end

end

---------------------------------------------------------------------------------------------------
--	SRNAnubHealerMod:FShouldAnnounce
--
--		Computes whether or not the user should announce blame
--
function SRNAnubHealerMod:FShouldAnnounce()

	return SRNAnubHealerMod.db.profile.fEnableBlame and
		(not SRNAnubHealerMod.db.profile.fBlameOnlyWhenAssist or IsRaidLeader() or IsRaidOfficer());

end

---------------------------------------------------------------------------------------------------
--	SRNAnubHealerMod:ZONE_CHANGED_NEW_AREA
--
--		Handler for zone change events.
--
function SRNAnubHealerMod:ZONE_CHANGED_NEW_AREA()
	SRNAnubHealerMod:CheckActiveState();
end

---------------------------------------------------------------------------------------------------
--	SRNAnubHealerMod:COMBAT_LOG_EVENT_UNFILTERED
--
--		Handler for combat log events.
--
function SRNAnubHealerMod:COMBAT_LOG_EVENT_UNFILTERED(event, iTimestamp, strEventType, strSourceGuid, strSourceName, iSourceFlags, strDestGuid, strDestName, iDestFlags, varParam1, varParam2, varParam3, varParam4, varParam5, varParam6, ...)

	local tmNow = GetTime();

	-- Scan the raid every 0.5 seconds
	if (tmNow - SRNAnubHealerMod.tmLastScan > 0.5) then
		local iMinRaidHP = 100000;
		local strMinRaidName = "error";
		local iSecondsLeft = 0;

		SRNAnubHealerMod.tmLastScan = tmNow;

		for i = 1,MAX_RAID_MEMBERS do
			local strUnitId = "raid" .. i;

			-- Remember the target's raid icon
			SRNAnubHealerMod.rgiRaidIcon[i] = GetRaidTargetIndex(strUnitId);
			if (SRNAnubHealerMod.rgiRaidIcon[i]) then
				SRNAnubHealerMod.rgiRaidIconLast[i] = SRNAnubHealerMod.rgiRaidIcon[i];
				SRNAnubHealerMod.rgtmRaidIconLastTime[i] = tmNow;
			end

			-- Only look at the unit's health under certain conditions
			if (not UnitIsDeadOrGhost(strUnitId) and UnitIsConnected(strUnitId)) then
				local _, _, _, _, _, _, strZone = GetRaidRosterInfo(i);

				if ("Trial of the Crusader" == strZone
						--or "Icecrown" == strZone -- DEBUG
						) then
					local iUnitHP = UnitHealth(strUnitId);

					if (iUnitHP > 0 and iUnitHP < iMinRaidHP) then
						iMinRaidHP = iUnitHP;
						strMinRaidName = UnitName(strUnitId);
					end
				end
			end
		end

		for i = s_iMaxComputedSecond,1,-1 do
			if (iMinRaidHP >= s_rgiHPPerSecond[i]) then
				iSecondsLeft = i;
				break;
			end
		end

		-- Update the frame, if necessary
		if (SRNAnubHealerMod.Frame and SRNAnubHealerMod.Frame:IsShown()) then
			local colorBG = s_colorDefaultBG;
			local strMsg;

			if (iSecondsLeft <= 3) then
				colorBG = s_colorRedBG;
			elseif (iSecondsLeft <= 6) then
				colorBG = s_colorYellowBG;
			end

			if (iSecondsLeft >= s_iMaxComputedSecond) then
				strMsg = ">15s"
			else
				strMsg = iSecondsLeft .. "s (" .. strMinRaidName .. ")";
			end

			SRNAnubHealerMod.FrameMessage:SetText(strMsg);
			SRNAnubHealerMod.Frame:SetBackdropColor(colorBG.r, colorBG.g, colorBG.b);
		end

		-- Play a sound, if necessary
		if (SRNAnubHealerMod.db.profile.fEnableLowHeathSound
				and iSecondsLeft <= SRNAnubHealerMod.db.profile.iLowHealthSoundLevel
				and tmNow - SRNAnubHealerMod.tmLastLeech < 5
				and tmNow - SRNAnubHealerMod.tmLastKlaxon > 1.9) then
			PlaySoundFile(s_strLowHeathWavPath);
			SRNAnubHealerMod.tmLastKlaxon = tmNow;
		end

	end

	-- Track Phase 3
	if ("SPELL_DAMAGE" == strEventType and "Leeching Swarm" == varParam2
		--or "SPELL_PERIODIC_HEAL" == strEventType and "Renew" == varParam2 -- DEBUG
		) then

		-- Announce the start of phase 3
		if (tmNow - SRNAnubHealerMod.tmLastLeech > 30) then
			SRNAnubHealerMod.tmPhase3Start = tmNow;

			if (SRNAnubHealerMod:FShouldAnnounce()) then
				SRNAnubHealerMod:AnnounceMessage("SRNAnubHealerMod: Starting phase 3");
			end
		end

		SRNAnubHealerMod.tmLastLeech = tmNow;

	-- Analyze actions on friendly players during phase 3
	elseif (tmNow - SRNAnubHealerMod.tmLastLeech < 10
		and 0 ~= bit.band(iDestFlags, COMBATLOG_OBJECT_TYPE_PLAYER)
		and 0 ~= bit.band(iDestFlags, COMBATLOG_OBJECT_REACTION_FRIENDLY)) then

		-- Blame player deaths
		if ("UNIT_DIED" == strEventType) then

			if (SRNAnubHealerMod:FShouldAnnounce()) then
				local iRaidIcon;
				local strAlternateCause;
				local fResidual = false;

				-- Look for the dead unit in the raid
				for i = 1,MAX_RAID_MEMBERS do
					if (strDestGuid == UnitGUID("raid" .. i)) then
						if (SRNAnubHealerMod.rgiRaidIcon[i]) then
							iRaidIcon = SRNAnubHealerMod.rgiRaidIcon[i];
						elseif (SRNAnubHealerMod.rgiRaidIconLast[i] and tmNow - SRNAnubHealerMod.rgtmRaidIconLastTime[i] < 6) then
							iRaidIcon = SRNAnubHealerMod.rgiRaidIconLast[i];
							fResidual = true;
						end

						if (SRNAnubHealerMod.rgsAlternateCauses[i] and tmNow - SRNAnubHealerMod.rgsAlternateCauses[i].tm < 3) then
							strAlternateCause = SRNAnubHealerMod.rgsAlternateCauses[i].str
						end
					end
				end

				-- If the dead unit had an icon, announce the blame
				if (iRaidIcon) then
					local iTimeDelta = floor(tmNow - SRNAnubHealerMod.tmPhase3Start);
					local strMsg = "T+" .. iTimeDelta .. ": " .. strDestName .. " died with icon " ..  s_rgstrIconEncodings[iRaidIcon];

					if (strAlternateCause) then
						strMsg = strMsg .. " (additional damage from " .. strAlternateCause .. ")";
					elseif (fResidual) then
						strMsg = strMsg .. " (probably from residual ticks of Penetrating Cold)";
					end

					SRNAnubHealerMod:AnnounceMessage(strMsg);
				end
			end

		-- Make a sound when PC is cast in Phase 3
		elseif ("SPELL_CAST_SUCCESS" == strEventType and "Penetrating Cold" == strEventType) then

			if (SRNAnubHealerMod.db.profile.fEnablePCSound) then
				PlaySoundFile(s_strPenetratingColdWavPath);
			end

		-- Track non-Penetrating Cold spell damage to players
		elseif (("SPELL_DAMAGE" == strEventType or "SPELL_PERIODIC_DAMAGE" == strEventType)
				and "Penetrating Cold" ~= varParam2) then

			-- Look for the victim in the raid and record the alternate cause of death
			for i = 1,MAX_RAID_MEMBERS do
				if (strDestGuid == UnitGUID("raid" .. i)) then
					SRNAnubHealerMod.rgsAlternateCauses[i] =
					{
						tm = tmNow,
						str = varParam2,
					};
				end
			end

		-- Track swing damage to players
		elseif ("SWING_DAMAGE" == strEventType) then

			-- Look for the victim in the raid and record the alternate cause of death
			for i = 1,MAX_RAID_MEMBERS do
				if (strDestGuid == UnitGUID("raid" .. i)) then
					SRNAnubHealerMod.rgsAlternateCauses[i] =
					{
						tm = tmNow,
						str = "Melee by " .. strSourceName,
					};
				end
			end

		end

	end

end

